//#define DEBUG

/*
Ver. 2.0  - 2018-02-17 (flashlighs by MaXyM)
Ver. 1.01 - 2015-06-17

This sketch emulates Thrustmaster TX RW 458 Italia wheel
allowing to connect arduino to TX RW wheelbase (may work for T500/T300 as well)
and emulate button presses - up to 17 buttons can be connected to arduino, if you
don't need a serial debugging

!!! There are 2 flavours of arduino boards - powered by +5 volts or +3.3 volts !!!
!!! If you have +5V Arduino version - disconnect it from USB before connecting to a wheelbase !!!
!!! Just power your arduino from TX RW wheelbase Black pin - it supplies +3.3V voltage (should be enough) !!!
!!! don't connect USB +5V wire to TX RW wheelbase Black +3.3V - you'll get a collision !!!

Thrustmaster Wheelbase cable pinout (PS/2 connector, also known as mini-Din 6pin female):
1 - Green    - not used
2 - Blue     - GND (ground)
3 - White    - MISO (master in, slave out - to read the data from the wheel)
4 - Orange   - SS (slave select, or PL - parallel load, set it to 0 when you want to read the data)
5 - Red      - CLK (clock impulses)
6 - Black    - +VCC (+3.3 volts! if you have Arduino 5V version - use external +3.3V power or use the wheelbase power!!!)

Arduino UNO pins -> TX RW wheelbase cable pins

Arduino GND                 -> TWRX Blue wire (2)
Arduino pin 12              -> TXRW White wire (3) (data from wheel to base)
Arduino pin 10 + pin 2 (SS) -> TXRW Orange wire pin (4) (yes, on Arduino it's wired to two pins. 10 - SS, 2 - INT0)
Arduino pin 13 (SCK)        -> TXRW Red wire (5)
Arduino +5V                 -> TXRW Black wire (6) (it gives 3.3V, but must be connected into +5V socket on arduino uno side)

Button mappings (we send 8 bytes to wheelbase, only 3 first bytes matter, only 17 bits are used for buttons)

Byte 1
  1 - not used
  1 - n/u
  0 - n/u
  0 - n/u
  0 - n/u
  0 - n/u
  0 - n/u
  1 – Gear Up (button 2 = R_Pad) -> Arduino digital pin 8 (portB bit0)
  
Byte 2
  1 – A (button 6)                  -> pin 7 (portD bit7)
  1 – B (button 4)                  -> pin 6 (portD bit6)
  1 – RS (button 12)                -> pin 5 (portD bit5)
  1 – Menu (button 9)               -> pin 4 (portD bit4)
  1 – Gear Down (button 1 = L_Pad)  -> pin 3 (portD bit3)
  1 – X (button 5)                  -> pin 9 (portB bit1)
  1 – Manettino CCW (button 7, View) -> pin 1 (portD bit1) - if you don't use serial debug (UART TX)
  1 – Manettino CW (button 8, Menu) -> pin 0 (portD bit0) - if you don't use serial debug (UART RX)

Byte 3
  1 – Suspend (button 13)           -> hey, not enough free pins on arduino uno left..
  1 – Y (button 3)                  -> pin 11 (portB bit3)
  1 – LS (button 11)                -> pin A5 (portC bit5)
  1 – View (button 10 – under X)    -> pin A4 (portC bit4)
  1 – D-Pad Down                    -> pin A3 (portC bit3)
  1 - D-Pad Right                   -> pin A2 (portC bit2)
  1 - D-Pad Left                    -> pin A1 (portC bit1)
  1 - D-Pad Up                      -> pin A0 (portC bit0)
  
Ah, each push button has 2 pins - you connect one to Arduino, another to the Ground :)
Pressed button gives you "0" (grounded signal), released = "1" (pulled-up to +VCC)
  
Free free to modify and use for your needs.
This sketch and the documentation above provided "AS IS" under the BSD New license.
http://opensource.org/licenses/BSD-3-Clause
(c) Taras Ivaniukovich (blog@rr-m.org) April 2015

http://rr-m.org/blog/hacking-a-thrustmaster-tx-rw-wheelbase-with-arduino-uno-part-2/
*/


/*
 Flashing headlighs
 Since FIA GT allows to flash headlights for various reasons, and almost no sim gives to platyer
 such feature, here is: hardware flashing lights. It works regardless the headlights are on or off. 
 It always generates multiply of two impulses to button mapped as headlights.

 Please note, we have to sacrifice 1 input for flash-button. Because of this, we end-up with 16 buttons
 but only 15 can be mapped to in-game functions.
 
 For sake of simplicity PortD is used. it mapps the most inputs (out of 3 ports used).
 We have to select 2 inputs: 
 1. for flasher button
 2. for headlighs

 Headlights button is one which in game is mapped to headlights. This function remains unchanged.
 Input dedicated to flashlights, is masked from data being sent to the wheel.

 There are some configuration parameters which affect flashlighs behaviour. See FLASH LIGHTS CONFIG.
 Note, that game is not as fast as controller, also probably intentionaly applying some input filtering.
 Therefore it might not react on too quick changes of input state. This is the reason why "sustain" is implemented.
 Also, going for too low fill_factor might cause the game being lost counting impulses. In result headlights might flash irregularily 
 and/or not returning to initial state of headlights. Default values are initial ones I managed to work with ProjectCars2. 
 However it might require some tuning.

*/


//#include "pins_arduino.h" // not necessary in 1.6.3?

byte wheelState [8];
volatile byte pos;


/**** FLASH LIGHTS CONFIG ****/

const long    interval        = 140; // interval at which to flash (milliseconds)
const float   fill_factor     = 0.5; // ratio of how long the lights remain lit, relaive to interval
const int     sustain         = 20;  // controller is able to scan single impulses, but games are slower. Simulation of some sustain is needed
const int     flashes         = 5;   // how many times to flash on single button press

// define bits being used as flash button and light button
// I have chosen portD (most inputs in a byte)
// In case you want to use another bits, remember bit 2 (counted from 0) of this port cannot be used
const byte flash_mask = B00000001;  // TX pin (Digital pin 1) for flash button
const byte lighs_mask = B10000000;  // RX pin (Digital pin 0) for lights button


/**** FLASH LIGHTS VARIABLES ****/

unsigned long previousMillis  = 0; // reference time for frequency measurment
int           tmpcount        = 0;
byte          flash           = 0; // variable for handling single click
int           flashcount      = 0; // variable for counting flashes (we need to always flash at least 2 times or multiply of 2)


void setup (void) {    
  DDRB  |= B00001011; // digital pins 8,9,11 used as inputs with a pull-up to +VCC
  PORTB |= B00001011;
  
  DDRC  |= B00111111; // pins 14-19 (A0 - A5) also used as digital inputs
  PORTC |= B00111111; // pulled-up to +VCC via internal 100k resistors
  
  DDRD  |= B11111011; // digital pins 0,1,3,4,5,6,7 used as inputs
  PORTD |= B11111011; // pulled-up to +VCC via internal 100k resistors
  
  wheelState[0] = B11000001; // TX RW Ferrari 458 Italia Wheel first data byte
  wheelState[1] = B11111111; // second data byte - buttons
  wheelState[2] = B11111111; // third data byte - buttons
  wheelState[3] = B11111111; // this and below - not used, but wheelbase reads all 8 bytes...
  wheelState[4] = B11111111;
  wheelState[5] = B11111111;
  wheelState[6] = B11111111;
  wheelState[7] = B11111111;

#ifdef DEBUG
  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, LOW);
  Serial.begin(9600);    // Arduino debug console - occupies pins RX (0) and TX (1) on Uno
#else
  
  pinMode(MISO, OUTPUT); // arduino is a slave device
  SPCR |= _BV(SPE);      // Enables the SPI when 1
  SPCR |= _BV(SPIE);     // Enables the SPI interrupt when 1

  // interrupt for SS rising edge. Arduino Uno Pin10 must be connected to Pin2!!!
  attachInterrupt (0, ss_rising, RISING);
#endif

}

// Interrupt0 (external, pin 2) - prepare to start the transfer
void ss_rising () {
  SPDR = wheelState[0]; // load first byte into SPI data register
  pos = 1;
}

// SPI interrupt routine
ISR (SPI_STC_vect) {
  SPDR = wheelState[pos++]; // load the next byte to SPI output register and return.
}

void loop() {
  #ifdef DEBUG
  delay(50);
  #endif
  
  flash              = 0;                           // release flash button
  bool repeat        = (flashcount % 2 > 0);        // evaluate modulo 2 condition for multiple use
  bool flash_pressed = ((PIND & flash_mask) == 0);  // evaluate flash button pressed for multiple use


  // go to period count
  // if flash button is pressed (pressed = zero), or
  // we need to complete second press (we need always 2 presses or its multiplicity
  if (flash_pressed || repeat || tmpcount > 0) {

    if ((tmpcount == 0) && flash_pressed) tmpcount = flashes;
    unsigned long currentMillis = millis();

    if (currentMillis - previousMillis >= (repeat ? interval * fill_factor : interval)) {

      previousMillis = currentMillis;

      flashcount++;
      if (!repeat) tmpcount--;
      flash = 1;

      #ifdef DEBUG
      digitalWrite(LED_BUILTIN, !digitalRead(LED_BUILTIN));
      #endif
    }
  } else {
    flashcount = 0;
  }



  // no we can press the lights button
  if (flash==1) 
  {
    wheelState[0] = (PINB & B00000001) | B11000000;                                               // take bit 0 from PORTB - TX RW byte1
    wheelState[1] = (((PINB & B00000010) << 1) | (PIND & B11111011) | flash_mask) & ~lighs_mask;  // take bit 1 from PORTB + the rest from PORTD B11111x11
                                                                                                  // addidionaly send headlighs button (flash) and mask using flash button
    wheelState[2] = ((PINB & B00001000) << 3) | (PINC & B00111111) | B10000000;                   // take bit 3 from PORTB + bits 0-5 from PORTC
    delay(sustain);
  }
  else
  {
     wheelState[0] = (PINB & B00000001) | B11000000;                             // take bit 0 from PORTB - TX RW byte1
     wheelState[1] = ((PINB & B00000010) << 1) | (PIND & B11111011) | flash_mask;// take bit 1 from PORTB + the rest from PORTD B11111x11
     wheelState[2] = ((PINB & B00001000) << 3) | (PINC & B00111111) | B10000000; // take bit 3 from PORTB + bits 0-5 from PORTC
  }
}


